"""
Specialized elements for interactive experiments.
"""

import time
from typing import Union

from pymongo.collection import ReturnDocument
from jinja2 import Environment, PackageLoader, Template
from alfred3.element.core import Element
from alfred3 import icon
from alfred3._helper import inherit_kwargs
from alfred3.data_manager import DataManager as dm

from .chat import ChatManager

jenv = Environment(loader=PackageLoader(__name__, "templates"))

@inherit_kwargs(exclude=["height"])
class Chat(Element):
    """
    Provides a chat window.

    This chat window can be used for real-time chat as well as for
    a kind of message-board.

    Args:
        chat_id (str): Unique identifier for the chat. Participants will
            be grouped together based on the chat id. For a group chat,
            it is sensible to use the group id here.
        
        nickname (str): Every chat participant gets a nickname that
            will be used to identify participant's messages. You can
            choose a custom nickname for the current session here. Although
            advisable, it is not strictly necessary to use unique nicknames.
            By default, chat participants will be identified by 
            'Anonymous x', where x is generated by counting through chat
            participants.
        
        placeholder (str): Text to display as a placeholder in the input
            field. Defaults to '...'

        disabled (bool): If *True*, the input field and send button will
            be disabled. Participants will be able to read the chat, but
            not to post messages. Defaults to *False*.
        
        readonly (bool): If *True*, the input field and send button will
            be completely hidden. Participants will be able to read the 
            chat, but not to post messages. Defaults to *False*.

        button_text (str): Text to display on the send-button. Defaults
            to a 'paper-plane' icon.

        msg_width (str): Width of chat messages in percent of the total
            width of the chat window. Don't forget to use the '%' sign. 
            Defaults to '70%'.
        
        height (str): Element height. Supply a string with unit 
            understood by CSS. Defaults to '350px'.

        colors (dict): Dictionary of colors to use for identification
            of chat participants. The dict should have nicknames as keys
            and color values that can be understood by CSS as values.
            By default, the chat element will go through a predefined
            list of 11 colors.
        
        color_target (str): Defines, how colors are used to identify
            participants. Can be 'nickname', 'border', or 'none'.
            Defaults to 'nickname'.
        
        button_style (str): Can be used for quick color-styling, using
            Bootstraps default color keywords: btn-primary, btn-secondary,
            btn-success, btn-info, btn-warning, btn-danger, btn-light,
            btn-dark. You can also use the "outline" variant to get
            outlined buttons (eg. "btn-outline-secondary"). 
        
        background_color (str): Background color for the chat windows.
            Can be any color understood by CSS. Defaults to 'WhiteSmoke',
            a kind of light grey.
        
        allow_resize (bool): If *True*, participants can adjust the 
            height of the text input field. Defaults to True.
        
        refresh_interval (int, float): Time in seconds, determines how
            often the chat elements looks for new messages. Defaults to
            1.

        {kwargs}
    
    Examples:

        A minimal message-board style chat::

            import alfred3 as al
            import alfred3_interact as ali
            
            exp = al.Experiment()

            exp += al.Page(title="Chat Demo", name="chat_demo")
            exp.chat_demo += ali.Chat("test_chat")
        
        A group chat in an interactive experiment::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                exp.plugins.mm = ali.MatchMaker("a", "b", exp=exp)
            
            @exp.member
            class Match(ali.MatchingPage):

                def wait_for(self):
                    self.exp.plugins.group = self.exp.plugins.mm.match_groupwise()
                    return True
            
            @exp.member
            class Success(al.Page):
                title = "Match Successful!"

                def on_first_show(self):
                    self += self.exp.plugins.group.chat()

    """

    element_template = jenv.get_template("html/ChatElement.html.j2")
    js_template = jenv.get_template("js/chat_element.js.j2")

    def __init__(
        self,
        chat_id: str,
        nickname: str = None,
        placeholder: str = "Press enter to send your message, shift+enter for a linebreak",
        disabled: bool = False,
        readonly: bool = False,
        button_text: str = icon("paper-plane"),
        msg_width: str = "70%",
        height: str = "350px",
        colors: dict = None,
        color_target: str = "nickname",
        you_label: str = " (you)",
        button_style: str = "btn-dark",
        background_color: str = "WhiteSmoke",
        allow_resize: bool = True,
        refresh_interval: Union[int, float] = 1,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._chat_id = chat_id
        self.chat_manager = None
        self.nickname = nickname
        self.you_label = you_label

        self.msg_width = msg_width

        self.post_url = None
        self.load_url = None
        self.get_new_url = None
        self.get_all_url = None

        self.colors = colors
        self.color_target = color_target
        self.placeholder = placeholder
        self.button_text = button_text
        self.button_style = button_style
        self.height = height
        self.background_color = background_color
        self.disabled = True if readonly else disabled
        self.readonly = readonly
        self.allow_resize = allow_resize

        self._interval = refresh_interval

    @property
    def _js_data(self):
        d = {}
        d["post_url"] = self.post_url
        d["load_url"] = self.load_url
        d["get_new_url"] = self.get_new_url
        d["get_all_url"] = self.get_all_url
        d["interval"] = self._interval
        d["name"] = self.name
        d["own_nickname"] = self.chat_manager.nickname
        d["msg_width"] = self.msg_width
        d["color_target"] = self.color_target
        d["you_label"] = self.you_label
        return d

    def added_to_experiment(self, exp):
        super().added_to_experiment(exp)
        self.chat_manager = ChatManager(self.exp, self._chat_id, self.nickname, self.colors)

        self.post_url = self.exp.ui.add_callable(self.chat_manager.post_message)
        self.load_url = self.exp.ui.add_callable(self.chat_manager.load_messages)
        self.get_new_url = self.exp.ui.add_callable(self.chat_manager.get_new_messages)
        self.get_all_url = self.exp.ui.add_callable(self.chat_manager.get_all_messages)

        js = self.js_template.render(self._js_data)
        self.add_js(js)

        if self.readonly:
            css = f"#{self.name}-input-group {{display: none;}}"
            self.add_css(css)

    @property
    def template_data(self) -> dict:
        d = super().template_data
        d["placeholder"] = self.placeholder
        d["button_text"] = self.button_text
        d["button_style"] = self.button_style
        d["height"] = self.height
        d["background_color"] = self.background_color
        d["disabled"] = self.disabled
        d["allow_resize"] = self.allow_resize
        return d

class ViewMembers(Element):
    element_template = jenv.get_template("html/ViewMembersElement.html.j2")
    view_js = jenv.get_template("js/view_members.js.j2")
    
    def __init__(self, match_maker, refresh_interval: int = 10, **kwargs):
        super().__init__(**kwargs)
        self.match_maker = match_maker
        self.render_url = None
        self.refresh_interval = refresh_interval
    
    def added_to_experiment(self, exp):
        super().added_to_experiment(exp)
        
        self.render_url = self.exp.ui.add_callable(self.render_table_body)
        cols = ["Session", "Status", "Condition", "Last Page", "Start Day", "Start Time", "Last Move", "Group", "Role"]
        js = self.view_js.render(name=self.name, render_url=self.render_url, refresh_interval=self.refresh_interval, cols=cols)
        self.add_js(js)

    
    @property
    def template_data(self):
        d = super().template_data
        d["members"] = self.match_maker.member_manager.members()
        return d
    
    def render_table_body(self):

        members = self.match_maker.member_manager.members()
        tbody = []
        for m in members:
            last_page = m.last_page if len(m.last_page) < 20 else m.last_page[:17] + "..."

            d = {}
            d["Session"] = m.session_id[-4:]
            d["Status"] = m.status
            d["Condition"] = self._condition_of(m.session_id)
            d["Last Page"] = last_page
            d["Start Day"] = m.start_day
            d["Start Time"] = m.start_time
            d["Last Move"] = m.last_move
            d["Group"] = m.group_id[-4:] if m.group_id else "-"
            d["Role"] = m.role
            tbody.append(d)

        return {"data": tbody}
    
    def _condition_of(self, sid: str) -> str:
        q = {}
        q["exp_id"] = self.exp.exp_id
        q["type"] = dm.EXP_DATA
        q["exp_session_id"] = sid

        pro = {"exp_condition": True}
        doc = self.exp.db_main.find_one(q, projection=pro)
        return doc["exp_condition"]

class ToggleMatchMakerActivation(Element):
    element_template = jenv.get_template("html/ToggleMatchMakerActivation.html.j2")
    js_template = jenv.get_template("js/toggle_match_maker.js.j2")

    def __init__(self, match_maker, **kwargs):
        super().__init__(**kwargs)
        self.match_maker = match_maker
        self.toggle_url = None

    def added_to_experiment(self, exp):
        super().added_to_experiment(exp)
        self.toggle_url = self.exp.ui.add_callable(self.match_maker.toggle_activation)
        js = self.js_template.render(name=self.name, toggle_url=self.toggle_url)
        self.add_js(js)

    @property
    def template_data(self):
        d = super().template_data
        if self.match_maker.active:
            d["text"] = '<i class="fas fa-circle-notch fa-spin mr-2"></i>MatchMaker is ACTIVE'
            d["button_style"] = "btn-success"
        else:
            d["text"] = '<i class="fas fa-circle-notch mr-2"></i>MatchMaker is INACTIVE'
            d["button_style"] = "btn-secondary"
        return d

    


"""
Specialized pages for interactive experiments.
"""

import time
from abc import abstractmethod
from functools import wraps

from jinja2 import Template

import alfred3 as al
from alfred3.element.core import Element
from alfred3.element.misc import Callback, RepeatedCallback
from alfred3._helper import inherit_kwargs
from alfred3.exceptions import SessionTimeout

from .match import MatchMaker
from ._util import MatchingTimeout
from ._util import NoMatch
from .element import ViewMembers, ToggleMatchMakerActivation


class PasswordPage(al.WidePage):
    def __init__(self, password: str, match_maker_id: str, **kwargs):
        super().__init__(**kwargs)
        self.password = password
        self.match_maker_id = match_maker_id

    def on_exp_access(self):
        self += al.HideNavigation()
        self += al.Text(f"Matchmaker ID: {self.match_maker_id}", align="center")
        self += al.VerticalSpace("50px")
        self += al.PasswordEntry(
            toplab="Password", password=self.password, width="narrow", name="pw", align="center"
        )

        self += al.SubmittingButtons(
            "Submit", align="center", name="pw_submit", width="narrow", button_style="btn-primary"
        )

        # enables submit via enter-press for password field
        self += al.JavaScript(
            code="""$('#pw').on("keydown", function(event) {
        if (event.key == "Enter") {
            $("#alt-submit").attr("name", "move");
            $("#alt-submit").val("forward");
            $("#form").submit();
            }});"""
        )


class AdminPage(al.WidePage):
    def __init__(self, match_maker, **kwargs):
        super().__init__(**kwargs)
        self.match_maker = match_maker

    def on_exp_access(self):
        self += al.Text(f"Matchmaker ID: {self.match_maker.matchmaker_id}", align="center")
        self += ToggleMatchMakerActivation(match_maker=self.match_maker, align="center")
        self += al.VerticalSpace("30px")
        self += ViewMembers(match_maker=self.match_maker, name="view_mm")
        self += al.VerticalSpace("30px")
        self += al.Text(
            "Note: The MatchMaker Admin displays only sessions for which the matchmaking process has been started.",
            font_size="small",
            width="full",
        )

        self += al.HideNavigation()
        self += al.WebExitEnabler()
        self += al.Style(code=f"#view_mm {{font-size: 85%;}}")

        # datatables javascript package
        # self += al.Style(url="//cdn.datatables.net/1.10.24/css/jquery.dataTables.min.css")
        self += al.Style(
            url="https://cdn.datatables.net/1.10.24/css/dataTables.bootstrap4.min.css"
        )
        self += al.JavaScript(url="//cdn.datatables.net/1.10.24/js/jquery.dataTables.min.js")
        self += al.JavaScript(
            url="https://cdn.datatables.net/1.10.24/js/dataTables.bootstrap4.min.js"
        )


class DefaultWaitingTimeoutPage(al.Page):
    """
    Default page to be displayed upon waiting timeouts.
    """

    title = "Timeout"
    name = "default_waiting_timeout_page"

    def on_exp_access(self):
        self += al.VerticalSpace("50px")
        self += al.Html(al.icon("user-clock", size="80pt"), align="center")
        self += al.VerticalSpace("100px")
        self += al.Text("Sorry, waiting took too long.", align="center")


class DefaultWaitingExceptionPage(al.Page):
    """
    Default page to be displayed upon waiting exceptions.
    """
    title = "Experiment aborted"
    name = "default_waiting_exception_page"

    def on_exp_access(self):
        self += al.VerticalSpace("50px")
        self += al.Html(al.icon("user-clock", size="80pt"), align="center")
        self += al.VerticalSpace("100px")
        self += al.Text("Sorry, the experiment was aborted while waiting.", align="center")


@inherit_kwargs
class WaitingPage(al.NoNavigationPage):
    """
    A page that provides a waiting screen for synchronization in
    interactive experiments.

    This page is must be derived. You define the condition by defining
    the :meth:`.wait_for` method. The page has a default design, set in
    :meth:`.on_exp_access` - you can safely redefine this method to use
    your own design.

    Once the :meth:`.wait_for` method returns a *True*-like value (i.e.,
    a value for which ``bool(value) == True)`` holds, the
    WaitingPage will proceed to the next page automatically.

    .. important:: The :meth:`.wait_for` method *must* return a value,
        otherwise you will wait indefinitely.

    Args:
        wait_msg (str): Text to be displayed in the default layout.
            Defaults to None, in which case the text is
            "Waiting for other group members." Can be defined as a 
            class attribute.
        wait_timeout (int): Maximum waiting time in seconds. If *wait_for*
            does not return a *True*-like value within waiting time, the experiment
            session is aborted. Defaults to None, in which case the
            timeout is ``60 * 20``, i.e. 20 minutes.
            Can be defined as a class attribute.
        wait_sleep_time (int): Number of seconds in between two internal
            calls to :meth:`.wait_for`. Defaults to None, in which case
            a call will be made every two seconds.
            Can be defined as a class attribute.
        wait_timeout_page (alfred3.Page): A custom page to be displayed
            in case of a waiting timeout. Defaults to an instance of
            :class:`.DefaultWaitingTimeoutPage`.
            Can be defined as a class attribute.
        wait_exception_page (alfred3.Page): A custom page to be displayed
            in case of an exception during waiting. Defaults to an
            instance of :class:`.DefaultWaitingExceptionPage`.
            Can be defined as a class attribute.

        {kwargs}

    Examples:

        The example demonstrates how to use the waiting page in
        an ongoing experiment to wait until a group member has proceeded
        to a certain point in the experiment (more precisely, until a
        group member has filled a certain input element in this case)::


            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                mm = ali.MatchMaker("a", "b", exp=exp)
                exp.plugins.group = mm.match_stepwise()


            @exp.member
            class InputPage(al.Page):

                def on_exp_access(self):
                    self += al.TextEntry(leftlab="Enter text", name="el1", force_input=True)


            @exp.member
            class Sync1(ali.WaitingPage):

                def wait_for(self):
                    you = self.exp.plugins.group.you
                    return you.values.get("el1", False)

            exp += al.Page(title="Waiting successful", name="success")


    """

    #: str: Page title
    title = "Waiting"

    #: Text to be displayed in the default layout.
    #: Defaults to None, in which case the text is
    #: "Waiting for other group members."
    wait_msg: str = "Waiting for other group members."

    #: Maximum waiting time in seconds. If *wait_for*
    #: does not return a *True*-like value within waiting time, the experiment
    #: session is aborted. Defaults to None, in which case the
    #: timeout is ``60 * 20``, i.e. 20 minutes.
    wait_timeout: int = 60 * 20

    #: Number of seconds in between two internal
    #: calls to :meth:`.wait_for`. Defaults to None, in which case
    #: a call will be made every two seconds.
    wait_sleep_time: int = 2

    #: Abort page to be displayed on timeout
    wait_timeout_page = None

    #: Abort page to be displayed on other exceptions during waiting
    wait_exception_page = None

    #: If *True*, the page will save signals of activity in the database
    #: for this experiment session. Important for MatchMaking, but usually
    #: not for ordinary waiting. Defaults to *False*
    ping = False

    def __init__(
        self,
        *args,
        wait_msg: str = None,
        wait_timeout: int = None,
        wait_sleep_time: int = None,
        wait_timeout_page: al.Page = None,
        wait_exception_page: al.Page = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        if wait_msg is not None:
            self.wait_msg = wait_msg

        if wait_timeout is not None:
            self.wait_timeout = wait_timeout

        if wait_sleep_time is not None:
            self.wait_sleep_time = wait_sleep_time

        if wait_timeout_page is not None:
            self.wait_timeout_page = wait_timeout_page
        else:
            self.wait_timeout_page = DefaultWaitingTimeoutPage()

        if wait_exception_page is not None:
            self.wait_exception_page = wait_exception_page
        else:
            self.wait_exception_page = DefaultWaitingExceptionPage()
        
        self += RepeatedCallback(
            func=self._wait_for, 
            interval=self.wait_sleep_time, 
            followup="custom", 
            custom_js="if (data) {move('forward')};"
        )

        #: Time of waiting start in seconds since epoch
        #: We are not using the first show time here, because we want
        #: to try to synchronize the waiting timeout with the displayed
        #: Counter as best as possible. If we used the first show time,
        #: that may be an earlier time point.
        self.waiting_start: float = None
        self.countup = al.CountUp(font_size=30, align="center")

    
    @property
    def expiration_time(self) -> float:
        """
        float: Point in time at which the waiting page expires in 
        seconds since epoch.
        """
        if not self.waiting_start:
            return None
        
        return self.waiting_start + self.wait_timeout
    

    @property
    def expired(self) -> bool:
        """
        bool: *True* if the page has exceeded its maximum waiting time.
        """
        if self.expiration_time:
            now = time.time()
            return now > self.expiration_time
        else:
            return False
    
    
    @abstractmethod
    def wait_for(self):
        """
        Once this method returns *True*, the page automatically
        forwards participants to the next page.

        It will be repeatedly called internally with the time between
        two calls defined by :attr:`.wait_sleep_time`.
        """
        pass

    def _wait_for(self):
        """
        This method gets called repeatedly via ajax callback from the
        waiting page client-side. If it returns *True*, the experiment will move
        on to the next page. Otherwise, the callback will try again
        until the timeout is reached.
        """
        
        if self.ping:
            self._call_ping()
        
        if not self.waiting_start:
            self.waiting_start = time.time()
            self.countup.start_time = self.waiting_start
        
        if self.expired:
            self.log.exception("Timeout on waiting page. Aborting experiment.")
            self.exp.abort(reason=MatchMaker._TIMEOUT_MSG, page=self.wait_timeout_page)
            return True
        
        try:
            wait_status = self.wait_for()
        
        except NoMatch:
            return False # return False so that the repeated callback will try again
        
        except Exception:
            self.log.exception("Exception in waiting function. Aborting experiment.")
            self.exp.abort(reason="waiting error", page=self.wait_exception_page)
            return True
        
        return wait_status

    def on_exp_access(self):
        spinning_icon = al.icon("spinner", spin=True, size="90pt")

        self += al.VerticalSpace("100px")
        self += al.Text(spinning_icon, align="center")
        self += al.VerticalSpace("30px")
        self += self.countup
        self += al.Text(self.wait_msg, align="center")
    
    def _call_ping(self):
        sid = self.exp.session_id
        query = {
            "type": "match_maker",
            f"members.{sid}.session_id": sid,
        }
        update = {"members": {sid: {"ping": time.time()}}}
        self.exp.db_misc.find_one_and_update(query, update=[{"$set": update}])


@inherit_kwargs
class MatchingPage(WaitingPage):
    """
    A waiting page for matchmaking.

    Args:
        {kwargs}
    
    Examples:
        ::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                exp.plugins.mm = ali.MatchMaker("a", "b", exp=exp)
            
            @exp.member
            class Match(ali.MatchingPage):

                def wait_for(self):
                    self.exp.plugins.group = self.plugins.mm.match_groupwise()
                    return True

            @exp.member
            class Demo(al.Page):

                def on_first_show(self):
                    role = self.exp.plugins.group.me.role
                    self += al.Text(f"I was assigned to role '{{role}}'.")
    """
    #: If *True*, the page will save signals of activity in the database
    #: for this experiment session. Important for MatchMaking, but usually
    #: not for ordinary waiting. Defaults to *True*
    ping = True

    


"""
Functionality related to groups.
"""

import json
import random
import time
from pathlib import Path
from typing import Iterator
from dataclasses import asdict

from .member import MemberManager
from .member import GroupMember
from .data import GroupData
from .data import SharedGroupData
from ._util import MatchingError
from ._util import saving_method
from .element import Chat

class Group:
    """
    The group object holds members and keeps track of their roles.

    The group object's main task is providing access to the group members
    via their roles. Most importantly, the group offers the attributes
    :attr:`.me` and :attr:`.you`. Beyond that, you can use dot notation,
    using any member's role like an attribute on the group. The group will
    return a :class:`.GroupMember` object for the member inhabiting that
    role::

        import alfred3 as al
        from alfred3_interact import MatchMaker

        exp = al.Experiment()

        @exp.setup
        def setup(exp):
            mm = MatchMaker("a", "b", exp=exp)
            exp.plugins.group = mm.match_stepwise()
            print(exp.plugins.group.a) # Accessing the GroupMember with role "a"
        
        exp += al.Page(name="demo")

    See Also:
        For more details, refer to the :class:`.MatchMaker` documentation.

    """
    DATA_TYPE = "match_group"

    def __init__(self, matchmaker, **kwargs):
        kwargs.pop("_id", None)
        self.mm = matchmaker
        self.exp = self.mm.exp
        self.manager = MemberManager(matchmaker=self.mm)
        exp_id = kwargs.get("exp_id", None)
        kwargs["exp_id"] = exp_id if exp_id is not None else self.exp.exp_id
        mm_id = kwargs.get("matchmaker_id", None)
        kwargs["matchmaker_id"] = mm_id if mm_id is not None else self.mm.matchmaker_id
        if not "exp_version" in kwargs:
            kwargs["exp_version"] = self.mm.exp_version
        self.data = GroupData(**kwargs)
        self._operation_start = None

        self._shared_data = SharedGroupData(group=self)
        self._shared_data._fetch()
        self.exp.append_plugin_data_query(self._plugin_data_query)
        self._save()
    
    @property
    def _plugin_data_query(self):
        f = {"exp_id": self.exp.exp_id, "type": "match_group"}
        
        q = {}
        q["title"] = "Group Data"
        q["type"] = "match_group"
        q["query"] = {"filter": f}
        q["encrypted"] = False

        return q
    
    @property
    def shared_data(self) -> SharedGroupData:
        """
        SharedGroupData: Gives access to the group's shared data dictionary,
        which will always automatically synchronize data for all group 
        members.

        See Also:
            :class:`.SharedGroupData`
        """
        self._shared_data._fetch()
        return self._shared_data

    @property
    def path(self) -> Path:
        """
        Path: Path to group data file in offline group experiments.
        """
        parent = self.mm.io.path.parent
        return parent / f"group_{self.group_id}.json"

    @property
    def full(self) -> bool:
        """
        bool: Indicates whether the group is full or not. Counts only
        active members.
        """
        return len(list(self.active_or_finished_members())) == len(self.data.roles)

    @property
    def you(self) -> GroupMember:
        """
        GroupMember: :class:`.GroupMember`  object for
        the *other* participant in a dyad (i.e. a two-member-group).
        """
        if len(self.data.roles) > 2:
            raise MatchingError(
                "Can't use 'you' for groups with more than 2 slots, because it is ambiguous. Use the 'other_members()' generator instead, or access members based on their roles."
            )
        else:
            return next(self.other_members())

    @property
    def me(self) -> GroupMember:
        """
        GroupMember: :class:`.GroupMember` object for the own session.
        Useful to access the own role.
        """
        return next(m for m in self.members() if m.session_id == self.exp.session_id)
    
    def chat(self, **kwargs) -> Chat:
        """
        Shortcut for creating a group chat.

        Args:
            **kwargs: Any argument accepted by :class:`.Chat` can be used,
                including *chat_id* and *nickname*. Specifying the latter
                two will simply override the defaults.
        
        See Also:
            The basic functionality is provided in the :class:`.Chat`
            element, which can be used on its own.

        Returns:
            Chat: A chat element in which the default for the chat id is
            set to the group id and the default for the nickname is set 
            to each member's role. 
        """
        chat_id = kwargs.pop("chat_id", self.group_id)
        nickname = kwargs.pop("nickname", self.me.role)
        return Chat(chat_id=chat_id, nickname=nickname, **kwargs)

    def members(self) -> Iterator[GroupMember]:
        """
        Iterator: A generator, iterating over *all* members of the group.
        Yields :class:`.GroupMember` objects.

        See Also:
            - :meth:`.active_members`
            - :meth:`.other_members`
        """
        return (self.manager.find(mid) for mid in self.data.members)

    def active_members(self) -> Iterator[GroupMember]:
        """
        Iterator: A generator, iterating over all *active* members of the
        group. Yields :class:`.GroupMember` objects.

        See Also:
            - :attr:`.GroupMember.active`
        """
        return (member for member in self.members() if member.active)
    
    def active_or_finished_members(self) -> Iterator[GroupMember]:
        for member in self.members():
            if member.active or member.finished:
                yield member

    def other_members(self) -> Iterator[GroupMember]:
        """
        Iterator: A generator, iterating over the *active* group members
        except for :attr:`.me`. Yields :class:`.GroupMember` objects.
        """
        return (m for m in self.active_or_finished_members() if m.session_id != self.exp.session_id)

    def get_member_by_role(self, role: str) -> GroupMember:
        """
        GroupMember: Returns the :class:`.GroupMember` that inhabits the
        given role.
        """
        if not role in self.data.roles:
            raise AttributeError(f"Role '{role}' not found in {self}.")

        return next(m for m in self.active_members() if m.role == role)

    def get_member_by_id(self, id: str) -> GroupMember:
        """
        GroupMember: Returns the :class:`.GroupMember` belonging to the
        experiment session with session id *id*. 

        Raises:
            AttributeError: If there is no group member of the given id.
        """
        if not id in self.data.members:
            raise AttributeError(f"Member with id '{id}' not found in {self}.")

        return next(m for m in self.members() if m.data.session_id == id)

    def ongoing_roles(self) -> Iterator[str]:
        """
        Iterator: Iterates over roles for which sessions are currently 
        in progess.
        """
        for role, member_id in self.data.roles.items():
            if member_id is None:
                continue
            else:
                member = self.manager.find(id=member_id)
                if member.active:
                    yield role

    def open_roles(self) -> Iterator[str]:
        """
        Iterator: Iterates over open roles (str).
        """
        for role, member_id in self.data.roles.items():
            if member_id is None:
                yield role
            else:
                member = self.manager.find(id=member_id)
                if member.finished:
                    continue
                elif not member.active:
                    yield role

    def _shuffle_roles(self):
        role_list = list(self.data.roles.items())
        random.shuffle(role_list)
        self.data.roles = dict(role_list)

    def _assign_all_roles(self, to_members):
        if not self.full:
            raise MatchingError("Can't assign all roles if group is not full.")

        if not len(list(self.open_roles())) == len(self.data.roles):
            raise MatchingError("Some roles are already assigned, can't assign all roles.")

        for role, member in zip(self.data.roles.keys(), to_members):
            member.data.role = role
            self.data.roles[role] = member.data.session_id

    def _assign_next_role(self, to_member: GroupMember):
        role = next(self.open_roles())
        to_member.data.role = role
        self.data.roles[role] = to_member.data.session_id

    def _save(self):
        if saving_method(self.exp) == "local":
            self._save_local()
        elif saving_method(self.exp) == "mongo":
            self._save_mongo()

    def _save_local(self):
        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(asdict(self.data), f, indent=4)

    def _save_mongo(self):
        q = {}
        q["type"] = self.DATA_TYPE
        q["group_id"] = self.group_id
        data = self.exp.db_misc.find_one_and_replace(q, asdict(self.data), upsert=True)
        return data

    def __iadd__(self, member):
        # TODO: Add group ID to member's experiment data

        if member.data.session_id in self.data.members:
            return self

        member.data.group_id = self.data.group_id
        self.data.members.append(member.data.session_id)
        return self

    def __eq__(self, other):
        return (type(self) == type(other)) and (self.data.group_id == other.data.group_id)

    def __str__(self):
        roles = str(list(self.data.roles.keys()))
        nactive = len(list(self.active_members()))
        nroles = len(self.data.roles)

        return f"{type(self).__name__}(roles={roles}, {nactive}/{nroles} roles filled, id='{self.group_id[-4:]}')"

    def __repr__(self):
        return self.__str__()

    def __getattr__(self, name):
        try:
            return getattr(self.data, name)
        except AttributeError:
            return self.get_member_by_role(role=name)

    def __enter__(self):
        self.data.busy = True
        self._operation_start = time.time()
        self._save()  # returns data, if group was not busy before but is now busy
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        timeout = time.time() - self._operation_start > self.mm.group_timeout
        if timeout or exc_type:
            self.data.active = False
            self.exp.log.error(f"There was an error or timeout when operating on {self}. The \
                group was deactivated.")
        self.data.busy = False
        self._save()


class GroupManager:
    def __init__(self, matchmaker):
        self.mm = matchmaker
        self.exp = self.mm.exp

    @property
    def query(self):
        q = {}
        q["type"] = Group.DATA_TYPE
        q["matchmaker_id"] = self.mm.matchmaker_id
        q["exp_id"] = self.mm.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        return q

    @property
    def path(self):
        return self.mm.io.path.parent

    def groups(self) -> Iterator[Group]:
        if saving_method(self.exp) == "local":
            return self._local_groups()
        elif saving_method(self.exp) == "mongo":
            return self._mongo_groups()
        else:
            raise MatchingError("No saving method found. Try defining a saving agent.")

    def active(self):
        return (g for g in self.groups() if g.active)

    def notfull(self, ongoing_sessions_ok=True) -> Iterator[Group]:
        if ongoing_sessions_ok:
            for g in self.active():
                if not g.full:
                    yield g
        else:
            for g in self.active():
                if not g.full and not any(g.ongoing_roles()):
                    yield g

    def find(self, id: str):
        try:
            return next(g for g in self.groups() if g.data.group_id == id)
        except StopIteration:
            raise MatchingError("The group you are looking for was not found.")
    
    def _local_groups(self):
        for fpath in self.path.iterdir():
            if fpath.is_dir():
                continue

            elif not fpath.name.startswith("group"):
                continue

            else:
                with open(fpath, "r", encoding="utf-8") as f:
                    gdata = json.load(f)
                    if gdata["exp_version"] == self.mm.exp_version:
                        yield Group(matchmaker=self.mm, **gdata)

    def _mongo_groups(self):
        for doc in self.exp.db_misc.find(self.query):
            yield Group(matchmaker=self.mm, **doc)


"""
The under-the-hood administration of the chat.
"""

import time
import bleach
from pymongo.collection import ReturnDocument
from alfred3.exceptions import AlfredError


class ChatManager:
    """
    Manages a chat.

    Args:
        exp (alfred.experiment.ExperimentSession): The experiment session
            to which the chat belongs.
        chat_id (str): Unique identifier for the chat.
        nickname (str): Nickname for the identification of the current
            session.
        colors (dict): Dictionary of colors for identifying participants
            with color.
        encrypt (bool): If True, the content of the messages will be
            encrypted before saving to the database. Requires encryption
            to be enabled on the experiment. Defaults to True.
        ignore_aborted_sessions (bool): If True, the ChatManager will not
            output messages of aborted or expired sessions. Can be
            necessary in experiments with asynchronous interaction to
            prevent confusing chats. Defaults to True.

    """

    # colors from https://colorbrewer2.org/#type=qualitative&scheme=Paired&n=12
    DEFAULT_COLORS = [
        "#1f78b4",
        "#33a02c",
        "#e31a1c",
        "#ff7f00",
        "#6a3d9a",
        "#b15928",
        "#a6cee3",
        "#b2df8a",
        "#fb9a99",
        "#fdbf6f",
        "#cab2d6",
    ]

    def __init__(
        self,
        exp,
        chat_id: str,
        nickname: str = None,
        colors: dict = None,
        encrypt: bool = True,
        ignore_aborted_sessions: bool = True,
    ):
        self.exp = exp
        self.chat_id = chat_id
        self.colors = colors
        self.encrypt = encrypt
        self.ignore_aborted_sessions = ignore_aborted_sessions

        self._query = {}
        self._query["exp_id"] = self.exp.exp_id
        self._query["type"] = "chat_data"
        self._query["chat_id"] = self.chat_id

        self.nickname = nickname
        self.member_number = self._register_session()
        if self.nickname is None:
            self.nickname = "Anonymous " + str(self.member_number)

        self.data = None
        self._loaded_index = 0
        self._local_change_counter = 0
        self.color = self._find_color()

        self._inactive_sids = []
        self.exp.append_plugin_data_query(self._plugin_data_query)
    
    @property
    def _plugin_data_query(self):
        f = {"exp_id": self.exp.exp_id, "type": "chat_data"}
        
        q = {}
        q["title"] = "Chat"
        q["type"] = "chat_data"
        q["query"] = {"filter": f}
        q["encrypted"] = True

        return q

    def _find_color_index(self, n) -> int:
        n_colors = len(self.DEFAULT_COLORS)
        if n <= n_colors:
            return n

        else:
            while n > n_colors:
                n -= n_colors
            return n

    def _find_color(self):
        if self.colors:
            return self.colors[self.nickname]
        else:
            i = self._find_color_index(self.member_number)
            return self.DEFAULT_COLORS[i - 1]

    def _register_session(self) -> int:
        """Returns chat member number (a simple count)"""
        doc = self.exp.db_misc.find_one_and_update(
            self._query,
            update=[{"$set": {"sessions": {self.exp.session_id: "registered"}}}],
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        return len(doc["sessions"])

    def post_message(self, msg: str):
        """
        Posts a new message to the database.
        """
        if not msg:
            return

        msg = bleach.clean(msg)  # sanitize input

        if self.encrypt:
            msg = self.exp.encrypt(msg)

        msg_data = {}
        msg_data["sender_session_id"] = self.exp.session_id
        msg_data["timestamp"] = time.time()
        msg_data["msg"] = msg
        msg_data["nickname"] = self.nickname
        msg_data["color"] = self.color

        self.exp.db_misc.find_one_and_update(
            self._query, update={"$push": {"messages": msg_data}, "$inc": {"change_counter": 1}}, upsert=True
        )

    def load_messages(self) -> str:
        """
        Loads new messages from the database into the ChatManager instance.

        Returns:
            str: A status indicator. "pass" means that no new messages 
            have been found, "update" means that the internal message 
            storage has been updated.
        """
        data = self.exp.db_misc.find_one(self._query, projection={"change_counter": True})
        
        if data.get("change_counter", False) == self._local_change_counter:
            return "pass"

        self._local_change_counter = data["change_counter"]
        chat_data = self.exp.db_misc.find_one(self._query)

        if self.encrypt:
            for msg in chat_data.get("messages", []):
                msg["msg"] = self.exp.decrypt(msg["msg"])

        self.data = chat_data
        if self.ignore_aborted_sessions:
                self._update_session_status()
        
        return "update"


    def get_new_messages(self) -> tuple:
        """
        Returns:
            tuple: All new messages belonging to the chat
        """
        if self.data and self.data.get("messages", False):
            i, self._loaded_index = self._loaded_index, len(self.data["messages"])

            msgs = self.data["messages"][i:self._loaded_index]
            out_messages = [
                msg for msg in msgs if msg["sender_session_id"] not in self._inactive_sids
            ]

            return tuple(out_messages)
        else:
            return tuple()

    def get_all_messages(self) -> tuple:
        """
        Returns:
            tuple: All messages belonging to the chat
        """
        if self.data and self.data.get("messages", False):
            msgs = self.data["messages"]
            self._loaded_index = len(msgs)

            out_messages = [
                msg for msg in msgs if msg["sender_session_id"] not in self._inactive_sids
            ]

            # if self.encrypt:
            #     for msg in out_messages:
            #         msg["msg"] = self.exp.decrypt(msg["msg"])

            return tuple(out_messages)
        else:
            return tuple()

    def _update_session_status(self):
        """
        Updates the list of inactive session IDs.
        """
        if not self.data.get("messages", False):
            return
        sids = [msg["sender_session_id"] for msg in self.data["messages"]]
        sids = set(sids)

        uncertain_sids = sids - set(self._inactive_sids)

        for sid in uncertain_sids:
            query = {"type": "exp_data", "exp_session_id": sid}
            sdata = self.exp.db_main.find_one(
                query, projection={"exp_aborted": True, "exp_start_time": True, "exp_finished": True}
            )

            if sdata["exp_aborted"]:
                self._inactive_sids.append(sid)

            elif self.exp.session_timeout and not sdata["exp_finished"]:
                expired = time.time() - sdata["exp_start_time"] > self.exp.session_timeout
                if expired:
                    self._inactive_sids.append(sid)


"""
Provides the main MatchMaker functionality.
"""

import time
import json
import copy
import string
import re
from dataclasses import asdict

from pymongo.collection import ReturnDocument
from alfred3.alfredlog import QueuedLoggingInterface
from alfred3.experiment import ExperimentSession
from alfred3.exceptions import SessionTimeout

from .group import Group
from .group import GroupManager
from .member import GroupMember
from .member import MemberManager
from .data import MatchMakerData
from ._util import saving_method
from ._util import MatchingError
from ._util import BusyGroup
from ._util import NoMatch

class MatchMakerIO:
    def __init__(self, matchmaker):
        self.mm = matchmaker
        self.path.parent.mkdir(exist_ok=True, parents=True)

    @property
    def db(self):
        return self.mm.exp.db_misc

    @property
    def path(self):
        p = self.mm.exp.config.get("interact", "path", fallback="save/interact")
        name = f"{self.mm.matchmaker_id}{self.mm.exp_version}.json"
        return self.mm.exp.subpath(p) / name

    @property
    def query(self):
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["exp_id"] = self.mm.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        q["matchmaker_id"] = self.mm.matchmaker_id
        return q

    def save(self, data: MatchMakerData):
        if saving_method(self.mm.exp) == "mongo":
            self._save_mongo(data)
        elif saving_method(self.mm.exp) == "local":
            self._save_local(data)
        else:
            raise MatchingError("No saving method found. Try defining a saving agent.")

    def load(self) -> MatchMakerData:
        """
        Loads MatchMakerData. If there is none, creates a MatchMakerData
        document.
        """
        if saving_method(self.mm.exp) == "mongo":
            return self._load_mongo()
        elif saving_method(self.mm.exp) == "local":
            return self._load_local()

    def load_markbusy(self) -> MatchMakerData:
        """
        Loads MatchMakerData and marks it as busy, if it is not busy
        already. In the latter case, returns None.
        """
        self.mm.busy = True
        if saving_method(self.mm.exp) == "mongo":
            return self._load_markbusy_mongo()
        elif saving_method(self.mm.exp) == "local":
            return self._load_markbusy_local()

    def release(self) -> MatchMakerData:
        """
        Releases MatchMakerData from a 'busy' state.
        Happens only in groupwise matching, so there is only a mongoDB
        version of this one.
        """
        q = copy.copy(self.query)
        q["busy"] = True
        self.db.find_one_and_update(q, {"$set": {"busy": False}})

    def _save_mongo(self, data: MatchMakerData):
        self.db.find_one_and_replace(self.query, asdict(data))

    def _save_local(self, data: MatchMakerData):
        with open(self.path, "w", encoding="utf-8") as f:
            json.dump(asdict(data), f, sort_keys=True, indent=4)

    def _load_mongo(self):

        insert = MatchMakerData(
            exp_id=self.mm.exp.exp_id,
            exp_version=self.mm.exp_version,
            matchmaker_id=self.mm.matchmaker_id,
            active=self.mm._active,
        )

        data = self.db.find_one_and_update(
            self.query,
            {"$setOnInsert": asdict(insert)},
            upsert=True,
            return_document=ReturnDocument.AFTER,
        )
        data.pop("_id", None)
        return MatchMakerData(**data)

    def _load_local(self) -> MatchMakerData:
        """
        If there is a matchmaker json file, this returns the data from
        that file. Otherwise, it inserts the needed json file and then
        returns the corresponding data.
        """
        if self.path.exists() and self.path.is_file():
            with open(self.path, "r", encoding="utf-8") as f:
                return MatchMakerData(**json.load(f))
        else:
            data = MatchMakerData(
                exp_id=self.mm.exp.exp_id,
                exp_version=self.mm.exp_version,
                matchmaker_id=self.mm.matchmaker_id,
                active=self.mm._active,
            )

            with open(self.path, "w", encoding="utf-8") as f:
                json.dump(asdict(data), f, sort_keys=True, indent=4)
            return data

    def _load_markbusy_local(self):
        data = self._load_local()
        if not data.busy:
            data.busy = True
            self._save_local(data)
            return MatchMakerData(**data)
        else:
            return None

    def _load_markbusy_mongo(self):
        q = self.query
        q["busy"] = False
        data = self.db.find_one_and_update(q, {"$set": {"busy": True}})
        if data is not None:
            data.pop("_id", None)
            data["busy"] = True
            return MatchMakerData(**data)
        else:
            return None
    

    def __enter__(self):
        return self.load_markbusy()

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            self.mm.member.data.active = False
            self.mm.member._save()
            self.exp.abort(reason="matchmaker_error")
            self.exp.log.error(f"There was an error in a locked MatchMaker operation. \
                I deactivated the responsible member {self.mm.member} and released the lock.")
        self.release()
        self.mm._data = self.load()



class MatchMaker:
    """
    Organizes the creation and composition of groups for interactive
    experiments.

    Args:
        *roles: A variable number of strings, indicating which roles
            will be appointed to members of the group. The number of
            roles determines the number of group members. All roles in a
            group must be unique. Roles cannot start with numbers or 
            contain spaces.
        exp (alfred3.experiment.ExperimentSession): The alfred3 experiment
            in which this matchmaker is used.
        admin_pw (str): Password for the MatchMaker admin view. If no
            admin password is supplied, the admin view is deactivated.
        id (str): Used in combination with the experiment ID as unique
            identifier for the MatchMaker.  Defaults to 'matchmaker',
            which is usually fine, unless you are using multiple
            matchmakers in one experiment.
        respect_version (bool): If True, the matchmaker will only match
            sessions that run on the same experiment version into the
            same group. This setting makes sure that there's no strange
            behavior if you make changes to an ongoing experiment.
            Defaults to True.
        active (bool): If True, the matchmaker will start in active mode.
            Defaults to *True*.
        inactive_page (Page): Page to be displayed to new participants
            if the MatchMaker is inactive. If None, a default page is
            used.
        admin_param (str): Name of the URL parameter used to start the
            admin mode. Defaults to the matchmaker id.


    Notes:
        See :meth:`.match_stepwise` and :meth:`.match_groupwise` for
        more information and examples. Also, there is an alternative
        constructor :meth:`.from_size`.

        .. note:: Note that the automatic MatchMaker admin mode is only
            available if the matchmaker is initialized in the experiment
            setup hook!

    See Also:
            - See :class:`.MatchingPage` for a special page class that
              offers a nice waiting screen and automatic forwarding upon
              achieving a match.
            
            - See :class:`.WaitingPage` for a special page class that
              offers a nice waiting screen for synchronization in an 
              ongoing experiment. This is useful to pause at some points
              in the experiment and wait for all group members to arrive
              at a specified point in the experiment.

    Examples:
        The example below illustrates the creation of an *asynchronous*
        group via :meth:`.match_stepwise`. For a synchronous group,
        refer to :meth:`.match_groupwise`::

            import alfred3 as al
            import alfred3_interact as ali

            exp = al.Experiment()

            @exp.setup
            def setup(exp):
                mm = ali.MatchMaker("role1", "role2", exp=exp)
                exp.plugins.group = mm.match_stepwise()
            
            
            @exp.member
            class Success(al.Page):
                title = "Match successful"

                def on_exp_access(self):
                    group = self.exp.plugins.group
                    
                    txt = f"You have successfully matched to role: {group.me.role}"
                    self += al.Text(txt, align="center")



    """

    _TIMEOUT_MSG = "MatchMaking timeout"
    _DATA_TYPE = "match_maker"

    def __init__(
        self,
        *roles,
        exp: ExperimentSession,
        id: str = "matchmaker",
        respect_version: bool = True,
        active: bool = True,
        inactive_page=None,
        admin_pw: str = None,
        admin_param: str = None,
    ):
        self.exp = exp
        
        if exp.start_time:
            raise MatchingError("MatchMaker must be initialized during experiment setup.")
        
        self.exp_version = self.exp.version if respect_version else ""
        self.log = QueuedLoggingInterface(base_logger="alfred3")
        self.log.add_queue_logger(self, __name__)
        self._active = active
        self.inactive_page = inactive_page
        self.admin_param = admin_param if admin_param is not None else id
        self.admin_pw = admin_pw
        self.admin_mode = False

        self.member_timeout = None

        #: Number of seconds after which a single
        #: group-operation will be considered to have failed. In this
        #: case, the group in question will be marked as inactive and
        #: not be included in further matchmaking. Defaults to 300 seconds
        #: (5 minutes).
        self.group_timeout = 10

        if len(roles) != len(set(roles)):
            raise ValueError("All roles in a group must be unique.")

        self.roles = self._validate_roles(roles)
        self.matchmaker_id = id
        self.io = MatchMakerIO(self)

        self._data = self.io.load()

        self.member_manager = MemberManager(self)
        self.group_manager = GroupManager(self)
        self.group = None
        self.member = None

        self.exp.abort_functions.append(self._deactivate_session)
        
        if self.admin_pw:
            self.admin_mode = self._enable_admin_mode(self.exp)
        
        
            

    @classmethod
    def from_size(cls, n: int, **kwargs):
        """
        Creates a matchmaker instance from the number of group members.

        Args:
            n (int): The number of group members.
            **kwargs: Further keyword arguments passed on to the usual
                constructor.

        Notes:
            Group roles will be the letters of the alphabet: First small
            letters, then capitalized, then repeated small letters
            (e.g. "aa"), and so on.
        """
        roles = []
        for i in range(n):
            letter = string.ascii_letters[i]
            while letter in roles:
                letter = letter + string.ascii_letters[i]
            roles.append(letter)

        return cls(*roles, **kwargs)

    @property
    def active(self) -> bool:
        """
        Returns *True*, if the MatchMaker is active.
        """
        d = self.io.load()
        self._active = d.active
        return self._active

    @property
    def member_timeout(self) -> int:
        """
        int: Timeout, after which members will be marked as inactive.
        """
        return self._member_timeout

    @member_timeout.setter
    def member_timeout(self, value: int):
        self._member_timeout = value if value is not None else self.exp.session_timeout

    @property
    def data(self) -> MatchMakerData:
        """
        MatchMakerData: The MatchMaker's underlying data.
        """
        self._data.members[self.member.session_id] = asdict(self.member.data)
        return self._data

    def match_stepwise(self, member_timeout: int = None, ongoing_sessions_ok: bool = False) -> Group:
        """
        Assigns participants to groups and roles one-by-one without
        waiting for a group to be full.

        Args:
            member_timeout (int): Number of seconds after which an experiment
                session is considered inactive. Inactive sessions are not
                included in the matchmaking process (but can in principle
                still finish their session). The roles of inactive sessions
                will be free for allocation to new members. If None,
                :attr:`alfred3.experiment.ExperimentSession.session_timeout`
                will be used. Defaults to None.
            ongoing_sessions_ok (bool): If False, new members will only
                be added to a group if all previous members of that group
                have finished their experiment sessions. This can prevent
                session losses in case 


        Roles are assigned to members in the order in which they
        are matched (first come, first serve).

        This method is the correct choice for groups that can operate
        in an asynchronous fashion. One example for such a setting is
        a "yoking" setup.

        Let's take a yoking group with two members, "a" and "b", as an
        example: The first participant starts her experiment
        and is immediately assinged to role "a". She finishes her
        experiment without needing access to any values from participant
        "b". The next participant is assigned to role "b". In his version
        of the experiment, some aspects depend on the inputs of
        participant "a".

        Examples:
            ::

                import alfred3 as al
                from alfred3_interact import MatchMaker

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    mm = MatchMaker("a", "b", exp=exp)
                    exp.plugins.group = mm.match_stepwise()

                @exp.member
                class Demo(al.Page):

                    def on_exp_access(self):
                        role = self.exp.plugins.group.me.role
                        self += al.Text(f"I was assigned to role '{role}'.")

        """
        if not self._check_activation():
            return

        self.member_timeout = member_timeout
        self.member = GroupMember(self)
        self.member._save()

        # match to existing group
        if any(self.group_manager.notfull(ongoing_sessions_ok=ongoing_sessions_ok)):
            try:
                self.group = self._match_next_group(ongoing_sessions_ok=ongoing_sessions_ok)
            except BusyGroup:
                self.group = self._wait_until_free(self.group_timeout)

            self.log.info(f"Session matched to role '{self.member.role}' in {self.group}.")
            self._save_infos()
            return self.group

        # start a new group
        else:
            roles = {role: None for role in self.roles}
            with Group(self, roles=roles) as group:
                self.log.info(f"Starting new group: {group}.")

                group += self.member
                group._assign_next_role(to_member=self.member)
                self.member._save()
                self.group = group
                self.log.info(f"Session matched to role '{self.member.role}' in {group}.")

            self._save_infos()
            return self.group

    def match_groupwise(
        self,
        ping_timeout: int = 15
    ) -> Group:
        """
        Waits until there are enough participants for a full group, then
        matches them together.

        Args:
            ping_timeout (int): Number of seconds after which an experiment
                session will be excluded from groupwise matching. This makes
                sure that only currently active sessions will be allocated
                to a group. Defaults to 15 (seconds).

        This method is the correct choice if group members exchange data
        in real time. Roles are assigned randomly.

        Notes:
            .. important:: Note that you must use a :class:`.MatchingPage` for 
                groupwise matching. Outside of a MatchingPage, groupwise
                matching will not work.

        See Also:
            - See :class:`.MatchingPage` for a special page class that
              offers a nice waiting screen and automatic forwarding upon
              achieving a match.
            
            - See :class:`.WaitingPage` for a special page class that
              offers a nice waiting screen for synchronization in an 
              ongoing experiment. This is useful to pause at some points
              in the experiment and wait for all group members to arrive
              at a specified point in the experiment.

        Examples:
            ::

                import alfred3 as al
                import alfred3_interact as ali

                exp = al.Experiment()

                @exp.setup
                def setup(exp):
                    exp.plugins.mm = ali.MatchMaker("a", "b", exp=exp)
                
                @exp.member
                class Match(ali.MatchingPage):

                    def wait_for(self):
                        self.exp.plugins.group = self.plugins.mm.match_groupwise()
                        return True

                @exp.member
                class Demo(al.Page):

                    def on_first_show(self):
                        role = self.exp.plugins.group.me.role
                        self += al.Text(f"I was assigned to role '{role}'.")

        """
        if not self._check_activation():
            return

        if not saving_method(self.exp) == "mongo":
            raise MatchingError("Must use a database for groupwise matching.")

        if self.member and self.group:
            self.log.info(f"match_groupwise was called, but {self.member} was already matched to \
                {self.group}. Returning group.")
            return self.group

        self._set_ping_timeout(ping_timeout)

        self.member = self._init_member()
        
        self.log.debug("Trying match")
        self.group = self._do_match_groupwise(ping_timeout=ping_timeout)

        if not self.group:
            self.log.debug("No match found.")
            raise NoMatch

        
        self.log.info(f"{self.group} filled in groupwise match.")
        self._save_infos()
        return self.group
    
    def _init_member(self):
        if self.member:
            return self.member
        
        member = GroupMember(self)
        member._save()
        if member.expired:
            raise SessionTimeout
        else:
            return member

    def toggle_activation(self) -> str:
        """
        Toggles MatchMaker activation.

        Returns:
            str: Returns the new status (active/inactive).
        """
        data = self.io.load()
        data.active = not data.active
        self.io.save(data=data)
        self._data = data

        return "active" if data.active else "inactive"

    def _wait_until_free(self, wait_max: int):
        if saving_method(self.exp) == "local":
            raise MatchingError("Can't wait for result in local experiment.")

        self.log.info(
            f"Waiting for ongoing group assignment to finish. Waiting for a maximum of {wait_max} seconds."
        )
        group = next(self.group_manager.notfull())
        start = time.time()
        timeout = False

        while group.busy and not timeout:
            timeout = time.time() - start > wait_max
            group = self.group_manager.find(group.group_id)
            time.sleep(1)

        if not timeout and not group.busy:
            self.log.info("Waiting successful. Proceeding with group assignment.")
            return group

        elif timeout:
            msg1 = f"{group} was in waiting position for too long. "
            msg2 = f"Starting a new group for session {self.exp.session_id}."
            self.log.warning(msg1 + msg2)

            roles = {role: None for role in self.roles}
            group = Group(self, roles=roles)
            return group

    def _match_next_group(self, ongoing_sessions_ok: bool):
        with next(self.group_manager.notfull(ongoing_sessions_ok=ongoing_sessions_ok)) as group:
            self.log.info(f"Starting stepwise match of session to existing group: {group}.")
            group += self.member
            group._assign_next_role(to_member=self.member)
            self.member._save()

            return group

    def _do_match_groupwise(self, ping_timeout):

        with self.io as data:
            if data is None:
                self.log.debug("Returning. Data marked, MM is busy.")
                return None

            self.member = self.member._load()
            if self.member.matched:
                self.log.debug("Returning. Found group.")
                return self.group_manager.find(self.member.data.group_id)

            waiting_members = self._get_waiting_members(ping_timeout)

            if len(waiting_members) >= len(self.roles):
                group = self._fill_group(data, waiting_members)
                return group
            
            self.log.debug("Returning. Not enough active members.")
            self.log.debug(f"Active members: {waiting_members}")
            return None
    
    def _get_waiting_members(self, ping_timeout):
        waiting_members = list(self.member_manager.waiting(ping_timeout=ping_timeout))
        self.log.debug(f"Full waiting members list: {waiting_members}")
        waiting_members = [m for m in waiting_members if m != self.member]
        self.log.debug(f"Filtered waiting members list: {waiting_members}")
        waiting_members.insert(0, self.member)
        self.log.debug(f"Changed waiting members list: {waiting_members}")
        return waiting_members
    
    def _fill_group(self, data, waiting_members):
        self.log.debug("Filling group.")
        roles = {role: None for role in self.roles}
        group = Group(matchmaker=self, roles=roles)

        candidates = (m for m in waiting_members)
        while not group.full:
            m = next(candidates)
            self.log.debug(f"Filling group with {m}")
            group += m

        group._assign_all_roles(to_members=waiting_members)
        group._save()

        # update matchmaker data
        for m in waiting_members:
            self.log.debug(f"Member {m} is being updated.")
            data.members[m.data.session_id] = asdict(m.data)
        self.io.save(data=data)
        self.log.debug(f"List of active members: {list(self.member_manager.active())}")

        self.log.debug("Returning filled group.")
        return group

    def _matching_timeout(self, group: Group, timeout_page):
        self.log.warning("Matchmaking timeout.")
        if group:
            group.data.active = False
            group._save()
            self.log.warning(f"{group} marked as expired.")
        self.io.release()
        self._data = self.io.load()

        if timeout_page:
            self.exp.abort(reason=self._TIMEOUT_MSG, page=timeout_page)
        else:
            self.exp.abort(
                reason=self._TIMEOUT_MSG,
                title="Timeout",
                msg="Sorry, the matchmaking process timed out.",
                icon="user-clock",
            )

        return None

    def _check_activation(self):
        if self.admin_mode:
            return True
        
        if self.active:
            return True

        self.log.info("MatchMaking session aborted (MatchMaker inactive).")
        self.exp._allow_append = True
        if self.inactive_page:
            self.exp.abort(reason="matchmaker_inactive", page=self.inactive_page)
        else:
            self.exp.abort(
                reason="matchmaker_inactive",
                title="MatchMaking inactive",
                msg="Sorry, the matchmaking process is currently inactive. Please try again later.",
                icon="user-times"
            )
        self.exp._allow_append = False

        return False

    def _save_infos(self):
        prefix = "interact"
        while prefix in self.exp.adata:
            prefix += "_"
        self.exp.adata[prefix] = {}
        self.exp.adata[prefix]["groupid"] = self.group.group_id
        self.exp.adata[prefix]["role"] = self.member.role

    def _deactivate_session(self, exp):
        # gets called when the experiment aborts!
        if self.member:
            self.member.data.active = False
            self.member._save()

    def _enable_admin_mode(self, exp) -> bool:
        """
        Returns *True* if admin mode is called, *False* otherwise.
        """
        from alfred3_interact.page import PasswordPage, AdminPage

        if exp.urlargs.get(self.admin_param, False) == "admin" and not exp.session_status == "admin":
            exp.session_timeout = None
            exp.config.read_dict({"data": {"save_data": False}})
            exp.config.read_dict({"layout": {"show_progress": False}})
            exp.session_status = "admin"

            exp._allow_append = True
            exp += PasswordPage(
                password=self.admin_pw,
                match_maker_id=self.matchmaker_id,
                title="MatchMaker Admin",
                name="_pw_admin_page",
            )
            exp += AdminPage(match_maker=self, title="MatchMaker Admin", name="_admin_page")
            exp._allow_append = False

            return True
        
        else:
            return False

    def _set_ping_timeout(self, ping_timeout):
        if not self._data.ping_timeout == ping_timeout:
            q = {"type": "match_maker", "exp_id": self.exp.exp_id}
            q["exp_version"] = self.exp_version
            q["matchmaker_id"] = self.matchmaker_id
            self.exp.db_misc.find_one_and_update(q, update={"$set": {"ping_timeout": ping_timeout}})
    
    def _validate_roles(self, roles):
        p = re.compile(r"^\d|\s")
        for role in roles:
            if p.search(role):
                raise ValueError(f"Error in role '{role}': Roles must not start with numbers \
                    and must not contain spaces.")
        
        return roles


    def __str__(self):
        return f"{type(self).__name__}(id='{self.matchmaker_id}', roles={str(self.roles)})"

    def __repr__(self):
        return self.__str__()



"""
Functionality related to group members.
"""

import time
import datetime
import json
from dataclasses import asdict

from pymongo.collection import ReturnDocument
from alfred3.data_manager import DataManager as dm

from .data import GroupMemberData
from ._util import saving_method
from ._util import MatchingError


class GroupMember:
    """
    The group member object grants access to a member's experiment data.


    The group member object's most important job is to provide easy
    access to the member's experiment data through the following
    attributes. They provide the same objects as the corresponding
    attributes of the :class:`alfred3.experiment.ExperimentSession`
    object:

    - :attr:`.values`
    - :attr:`.session_data`
    - :attr:`.metadata`
    - :attr:`.client_data`

    """

    def __init__(self, matchmaker, **kwargs):
        kwargs.pop("_id", None)
        self.mm = matchmaker
        self.exp = self.mm.exp
        exp_id = kwargs.get("exp_id", None)
        kwargs["exp_id"] = exp_id if exp_id is not None else self.exp.exp_id
        sid = kwargs.get("session_id", None)
        kwargs["session_id"] = sid if sid is not None else self.exp.session_id
        
        self.data = GroupMemberData(**kwargs)

    @property
    def _path(self):
        return self.mm.io.path
    
    @property
    def status(self) -> str:
        """
        str: Returns a string, indicating the member's status.
        """
        ping_timeout = self.mm.io.load().ping_timeout
        if self.finished:
            return "finished"
        elif self.session_data["exp_aborted"]:
            return "aborted"
        elif self.active:
            if self.group_id:
                return "matched"
            elif ping_timeout and time.time() - self.ping > ping_timeout:
                return "ping timeout"
            else:
                return "unmatched"
        elif self.expired:
            return "session timeout"
        else:
            return "pre-matching"
        
    @property
    def start_time(self) -> str:
        """
        str: Session start-time in human-readbale format (H:M:S)
        """
        return time.strftime("%H:%M:%S", time.localtime(self.timestamp))
    
    @property
    def start_day(self) -> str:
        """
        str: Session start-day in human-readbale format (yyyy-mm-dd)
        """
        date = datetime.date.fromtimestamp(self.timestamp)
        return date.isoformat()
    
    @property
    def last_move(self) -> str:
        """
        str: Time of last movement in human-readable format (H:M:S)
        """
        if self.move_history:
            t = self.move_history[-1]["hide_time"]
            if t:
                return time.strftime("%H:%M:%S", time.localtime(t))
            else:
                return None
        else:
            return None

    
    @property
    def last_page(self) -> str:
        """
        str: Returns the name of the member's current page.
        """
        try:
            return self.move_history[-1]["target_page"]
        except IndexError:
            return "-landing page-"

    @property
    def expired(self) -> bool:
        """
        bool: True, if the session is expired. False otherwise.
        """
        return time.time() - self.data.timestamp > self.mm.member_timeout

    @property
    def active(self) -> bool:
        """
        bool: True, if :attr:`.GroupMember.data.active` is True and the
        member has not timed out. A member is deactivated, e.g. when the
        experiment session of that member aborts.
        """
        return self.data.active and not self.expired and not self.finished
    
    @property
    def finished(self) -> bool:
        """
        bool: Indicates, whether the member's experiment session is finished.
        """
        return self.session_data["exp_finished"]

    @property
    def matched(self) -> bool:
        """
        bool: Indicates, whether the member has been added to a group.
        """
        return True if self.data.group_id is not None else False

    @property
    def values(self) -> dict:
        """
        dict: Flat dictionary of input element values.
        
        See Also:
            The dict works just like 
            :attr:`alfred3.experiment.ExperimentSession.values`. The keys are
            the names of input elements in the member's experiment session.
            The values are the user inputs. 
        """
        d = dm.flatten(self.session_data).items()
        return {k: v for k, v in d if k not in dm._metadata and k not in dm.client_data}

    @property
    def session_data(self) -> dict:
        """
        dict: Full dictionary of experiment session data.

        See Also:
            The dict works just like 
            :attr:`alfred3.experiment.ExperimentSession.session_data`.
        """
        if saving_method(self.exp) == "local":
            p = self.exp.config.get("local_saving_agent", "path")
            iterator = dm.iterate_local_data(dm.EXP_DATA, directory=self.exp.subpath(p))
            d = next((d for d in iterator if d["exp_session_id"] == self.session_id), None)
            return d
        elif saving_method(self.exp) == "mongo":
            query = {"exp_id": self.exp.exp_id, "exp_session_id": self.session_id, "type": dm.EXP_DATA}
            
            d = self.exp.db_main.find_one(query)

            if d is None:
                self.mm.log.warning(f"{self} did not find any session data. Trying to recover by waiting.")
                start = time.time()
                while d is None:
                    if time.time() - start > 10:
                        break
                    time.sleep(1)
                    d = self.exp.db_main.find_one(query)
            
            if not d:
                raise MatchingError(f"Could not find session data for {self}.")
            
            return d


    @property
    def move_history(self) -> dict:
        """
        dict: Dictionary of movement history.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.move_history`.
        """
        return self.session_data.get("exp_move_history", None)

    @property
    def metadata(self) -> dict:
        """
        dict: Dictionary of experiment metadata.

        See Also:
            The dict works just like
            :attr:`alfred3.experiment.ExperimentSession.metadata`.
        """
        data = dm.flatten(self.session_data).items()
        return {k: v for k, v in data if k in dm._metadata}

    @property
    def client_data(self) -> dict:
        """
        dict: Dictionary of client data.

        See Also:
            The dict works just like 
            :attr:`alfred3.experiment.ExperimentSession.client_data`.
        """

        data = dm.flatten(self.session_data).items()
        return {k: v for k, v in data if k in dm.client_data}

    
    def _save(self):
        if saving_method(self.exp) == "local":
            self._save_local()
        elif saving_method(self.exp) == "mongo":
            self._save_mongo()

    def _save_local(self):
        """
        Updates the entry for the member in the matchmaker data.
        """
        with open(self._path, "r", encoding="utf-8") as f:
            mm = json.load(f)

        mm["members"][self.session_id] = asdict(self.data)

        with open(self._path, "w", encoding="utf-8") as f:
            json.dump(mm, f, sort_keys=True, indent=4)

    def _save_mongo(self):
        """
        Updates only the entry for the member in the matchmaker data.
        """
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["matchmaker_id"] = self.mm.matchmaker_id
        q["exp_id"] = self.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        d = {"members": {self.session_id: asdict(self.data)}}
        mm = self.exp.db_misc.find_one_and_update(
            q, [{"$set": d}], return_document=ReturnDocument.AFTER
        )

        if not self.session_id in mm["members"]:
            raise MatchingError

    def _load(self):
        """
        Returns an updated version of self, if the corresponsing matchmaker
        is not busy.
        """
        if saving_method(self.exp) == "local":
            mm = self._load_local()
        elif saving_method(self.exp) == "mongo":
            mm = self._load_mongo()

        if mm is None:
            return None

        self.data = GroupMemberData(**mm["members"][self.session_id])
        return self

    def _load_local(self):
        with open(self._path, "r", encoding="utf-8") as f:
            return json.load(f)

    def _load_mongo(self):
        q = {}
        q["type"] = self.mm._DATA_TYPE
        q["matchmaker_id"] = self.mm.matchmaker_id
        q["exp_id"] = self.exp.exp_id
        q["exp_version"] = self.mm.exp_version
        return self.exp.db_misc.find_one(q)

    def __str__(self):
        gid = self.data.group_id[-4:] if self.group_id is not None else "-"
        return f"{type(self).__name__}(role='{self.data.role}', session_id='{self.data.session_id[-4:]}', group='{gid}', start='{self.start_time}')"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        c1 = type(self) == type(other)
        c2 = self.data.session_id == other.data.session_id

        return all([c1, c2])

    def __getattr__(self, name):
        return getattr(self.data, name)

class MemberManager:
    def __init__(self, matchmaker):
        self.mm = matchmaker

    def members(self):
        data = self.mm.io.load().members.values()
        for mdata in data:
            yield GroupMember(matchmaker=self.mm, **mdata)

    def active(self):
        return (m for m in self.members() if m.active)

    def unmatched(self):
        return (m for m in self.active() if not m.matched)
    
    def matched(self):
        return (m for m in self.members() if m.matched)

    def waiting(self, ping_timeout: int):
        now = time.time()
        for m in self.active():
            member_ping_timeout = now - m.ping > ping_timeout
            if not m.matched and not member_ping_timeout:
                yield m

    def find(self, id: str):
        return next((m for m in self.members() if m.data.session_id == id), None)



"""
Soome basic utilities.
"""


def saving_method(exp) -> bool:
    if not exp.secrets.getboolean("mongo_saving_agent", "use"):
        if exp.config.getboolean("local_saving_agent", "use"):
            return "local"
    elif exp.secrets.getboolean("mongo_saving_agent", "use"):
        return "mongo"
    else:
        return None


class MatchingTimeout(Exception):
    pass


class MatchingError(Exception):
    pass

class BusyGroup(Exception):
    pass

class NoMatch(Exception):
    pass


"""
Data representation classes for interactive components.
"""

import time
from uuid import uuid4
from collections import UserDict
from dataclasses import dataclass, field
from typing import Dict, List, Tuple

from alfred3.util import prefix_keys_safely
from pymongo.collection import ReturnDocument

from ._util import saving_method


class SharedGroupData(UserDict):
    """
    Shared group data dictionary.

    This dictionary can be used just like a normal dictionary. 
    Its benefit is that it synchronises data to a database. 
    This way, the data is always automatically shared between all group 
    members.
    """

    def __init__(self, *args, group, **kwargs):
        super().__init__(*args, **kwargs)
        self.group = group
        self._db = self.group.exp.db_misc

        if self.group.data.shared_data:
            self.data = self.group.data.shared_data
        
        self.data["__group_id"] = self.group.group_id
    
    def _push_remote(self):
        doc = self._db.find_one_and_update(
            filter={"group_id": self.group.group_id}, 
            update={"$set": {"shared_data": self.data}},
            projection={"shared_data": True, "_id": False},
            return_document=ReturnDocument.AFTER
            )
        if doc is not None:
            self.data = doc["shared_data"]
    
    def _fetch_remote(self):
        doc = self._db.find_one(
            filter={"group_id": self.group.group_id},
            projection={"shared_data": True, "_id": False},
            )
        if doc is not None:
            self.data = doc["shared_data"]
    
    def _push_local(self):
        self.group.data.shared_data = self.data
        self.group._save()
    
    def _fetch(self):
        if saving_method(self.group.exp) == "mongo":
            self._fetch_remote()
        
        self.data["__last_access"] = time.time()
        self._push()
    
    def _push(self):
        if saving_method(self.group.exp) == "mongo":
            self._push_remote()
        elif saving_method(self.group.exp) == "local":
            self._push_local()
    
    def last_change(self, format: str = "%Y-%m-%d, %X") -> str:
        return time.strftime(format, time.localtime(self.data["__last_change"]))
    
    def last_access(self, format: str = "%Y-%m-%d, %X") -> str:
        return time.strftime(format, time.localtime(self.data["__last_access"]))

    def __getitem__(self, key):
        self._fetch()
        return super().__getitem__(key)
    
    def __setitem__(self, key, item):
        self._fetch()
        super().__setitem__(key, item)
        self.data["__last_change"] = time.time()
        self._push()
    
    def __delitem__(self, key):
        self._fetch()
        super().__delitem__(key)
        self.data["__last_change"] = time.time()
        self._push()


@dataclass
class GroupMemberData:
    exp_id: str
    session_id: str
    group_id: str = None
    role: str = None
    timestamp: float = field(default_factory=time.time)
    ping: float = field(default_factory=time.time)
    active: bool = True
    type: str = "match_member"


@dataclass
class GroupData:
    exp_id: str
    exp_version: str
    matchmaker_id: str
    roles: dict
    group_id: str = field(default_factory=lambda: uuid4().hex)
    members: list = field(default_factory=list)
    timestamp: float = field(default_factory=time.time)
    active: bool = True
    busy: bool = False
    shared_data: dict = field(default_factory=dict)
    type: str = "match_group"


@dataclass
class MatchMakerData:
    exp_id: str
    exp_version: str
    matchmaker_id: str 
    members: dict = field(default_factory=dict)
    busy: bool = False
    active: bool = False
    ping_timeout: int = None
    type: str = "match_maker"





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

# Parse version from _version.py in package directory
# See https://packaging.python.org/guides/single-sourcing-package-version/#single-sourcing-the-version
version = {}
with open("src/alfred3_interact/_version.py") as f:
    exec(f.read(), version)

setuptools.setup(
    name="alfred3_interact",
    version=version["__version__"],
    author="Johannes Brachem, Christian Treffenstädt",
    author_email="brachem@psych.uni-goettingen.de",
    description="Components for interactive experiments in the alfred3 framework.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jobrachem/alfred3-interact",
    packages=setuptools.find_packages("src"),
    package_data={
        "alfred3_interact": [
            "templates/html/*",
            "templates/js/*",
        ]
    },
    package_dir={"": "src"},
    install_requires=[
        "alfred3>=2.0",
        "bleach>=3.2.1"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)
